@@ -16,17 +16,19 @@ use crate::config::BackupConfig;
 use crate::remotes::remote;
 use crate::services::service::Service;
 
-use job_scheduler::{Job, JobScheduler};
+use cron::Schedule;
 use regex::Regex;
 use std::fmt;
 use std::io;
 use std::path::{Path, PathBuf};
 use std::str::FromStr;
+use tokio_cron_scheduler::JobSchedulerError;
+use tokio_cron_scheduler::{Job, JobScheduler};
 
 use chrono::Weekday;
 use log::{error, info};
 
-use futures::executor;
+use uuid::Uuid;
 
 #[derive(Debug)]
 pub enum Error {
@@ -50,16 +52,16 @@ impl fmt::Display for Error {
 
 pub struct Backup {
     pub name: String,
-    pub what: Box<dyn Service>,
-    pub r#where: Box<dyn remote::Remote>,
+    pub what: Box<dyn Service + Send + Sync>,
+    pub r#where: Box<dyn remote::Remote + Send + Sync>,
     pub remote_path: PathBuf,
     pub when: String,
     pub compress: bool,
-    pub schedule: cron::Schedule,
+    pub schedule: Schedule,
     pub keep_last: Option<u32>,
 }
 
-impl Backup {
+impl<'a> Backup {
     fn get_hours_and_minutes(when: &str) -> Option<(i8, i8)> {
         let re = Regex::new(r"(\d{2}):(\d{2})").unwrap();
         let cap = re.captures(when)?;
@@ -71,6 +73,36 @@ impl Backup {
         None
     }
 
+    fn log_result(
+        result: Result<(), remote::Error>,
+        name: &str,
+        file: &Path,
+        remote_name: &str,
+        remote_path: &Path,
+        compress: bool,
+    ) {
+        if result.is_ok() {
+            info!(
+                "[{}] Successfully uploaded {} {}: {} to [{}] {}",
+                name,
+                if compress { " and compressed" } else { "" },
+                if file.is_dir() { "folder" } else { "file" },
+                file.display(),
+                remote_name,
+                remote_path.display(),
+            );
+        } else {
+            error!(
+                "[{}] Error during upload{} of {}: {}. Error: {}",
+                name,
+                if compress { " or compression" } else { "" },
+                if file.is_dir() { "folder" } else { "file" },
+                file.display(),
+                result.err().unwrap()
+            );
+        }
+    }
+
     fn parse_daily(input: &str) -> Result<String, Error> {
         // Daily 12:30
         let daily = "daily";
@@ -232,10 +264,10 @@ impl Backup {
             monthly.unwrap_err()
         )))
     }
-    pub fn new(
+    pub async fn new(
         name: &str,
-        remote: Box<dyn remote::Remote>,
-        service: Box<dyn Service>,
+        remote: Box<dyn remote::Remote + Send + Sync>,
+        service: Box<dyn Service + Send + Sync>,
         config: &BackupConfig,
     ) -> Result<Backup, Error> {
         let when_to_schedule = Backup::parse_when(&config.when);
@@ -265,212 +297,191 @@ impl Backup {
         })
     }
 
-    pub fn schedule(
-        self,
+    pub async fn schedule(
+        &'static self,
         scheduler: &mut JobScheduler,
         schedule: cron::Schedule,
-    ) -> Result<(), Error> {
-        let remote = self.r#where;
-        let mut service = self.what;
-        let compress = self.compress;
-        let name = self.name;
-        let remote_prefix = self.remote_path;
-        let keep_last = self.keep_last;
+    ) -> Result<Uuid, JobSchedulerError> {
+        scheduler.add(
+            Job::new_async(schedule.to_string().as_str(), move |_uuid, _js| {
+                Box::pin(async move {
+                    let remote = &self.r#where;
+                    let service = &self.what;
+                    let compress = self.compress;
+                    let name = self.name.clone();
+                    let remote_prefix = self.remote_path.clone();
+                    let keep_last = self.keep_last;
 
-        let log_result = |result: Result<(), remote::Error>,
-                          name: &str,
-                          file: &Path,
-                          remote_name: &str,
-                          remote_path: &Path,
-                          compress: bool| {
-            if result.is_ok() {
-                info!(
-                    "[{}] Successfully uploaded {} {}: {} to [{}] {}",
-                    name,
-                    if compress { " and compressed" } else { "" },
-                    if file.is_dir() { "folder" } else { "file" },
-                    file.display(),
-                    remote_name,
-                    remote_path.display(),
-                );
-            } else {
-                error!(
-                    "[{}] Error during upload{} of {}: {}. Error: {}",
-                    name,
-                    if compress { " or compression" } else { "" },
-                    if file.is_dir() { "folder" } else { "file" },
-                    file.display(),
-                    result.err().unwrap()
-                );
-            }
-        };
+                    // First call dump, to trigger the dump service if present
+                    info!("[{}] Calling dump...", &name);
+                    let dump = match service.dump().await {
+                        Err(error) => {
+                            error!("{}", Error::GeneralError(error));
+                            return;
+                        }
+                        Ok(dump) => dump,
+                    };
 
-        let job = Job::new(self.schedule, move || {
-            // First call dump, to trigger the dump service if present
-            info!("[{}] Calling dump...", name);
-            let dump = match service.dump() {
-                Err(error) => {
-                    error!("{}", Error::GeneralError(error));
-                    return;
-                }
-                Ok(dump) => dump,
-            };
+                    let path = dump.path.clone().unwrap_or_default();
+                    if path.exists() {
+                        // When dump goes out of scope, the dump is removed by Drop.
+                        info!("[{}] Dumped {}. Backing it up", name, path.display());
+                    }
 
-            let path = dump.path.clone().unwrap_or_default();
-            if path.exists() {
-                // When dump goes out of scope, the dump is removed by Drop.
-                info!("[{}] Dumped {}. Backing it up", name, path.display());
-            }
+                    // Then loop over all the dumped files and backup them as specified
+                    let mut local_files = service.list().await;
 
-            // Then loop over all the dumped files and backup them as specified
-            let mut local_files = service.list();
+                    // If the local_files list contains a single file, the upload should be in the form:
+                    // /remote/prefix/filename
+                    // even if the local file is in /local/path/in/folder/filename
+                    let mut single_file = local_files.len() <= 1;
 
-            // If the local_files list contains a single file, the upload should be in the form:
-            // /remote/prefix/filename
-            // even if the local file is in /local/path/in/folder/filename
-            let mut single_file = local_files.len() <= 1;
+                    // If the local_files list is a list of multiple files, we suppose these files all
+                    // share the same root. To find the root we can simply find the shortest string.
+                    // In this way, we can remove the "root prefix" and upload correctly.
+                    // From:
+                    // - /local/path/in/folder/A
+                    // - /local/path/in/folder/B
+                    // To
+                    // - /remote/prefix/A
+                    // - /remote/prefix/B
+                    let local_files_clone = local_files.clone();
+                    let mut local_prefix = local_files_clone
+                        .iter()
+                        .min_by(|a, b| a.cmp(b))
+                        .unwrap()
+                        .as_path();
 
-            // If the local_files list is a list of multiple files, we suppose these files all
-            // share the same root. To find the root we can simply find the shortest string.
-            // In this way, we can remove the "root prefix" and upload correctly.
-            // From:
-            // - /local/path/in/folder/A
-            // - /local/path/in/folder/B
-            // To
-            // - /remote/prefix/A
-            // - /remote/prefix/B
-            let local_files_clone = local_files.clone();
-            let mut local_prefix = local_files_clone
-                .iter()
-                .min_by(|a, b| a.cmp(b))
-                .unwrap()
-                .as_path();
+                    // The local_prefix found is:
+                    // In case of a folder: the shortest path inside the folder we want to backup.
+                    // In case of a file: the file itself.
 
-            // The local_prefix found is:
-            // In case of a folder: the shortest path inside the folder we want to backup.
-            // In case of a file: the file itself.
+                    // If is a folder, we of course don't want to consider this a prefix, but its parent.
+                    if !single_file {
+                        local_prefix = local_prefix.parent().unwrap();
+                    }
 
-            // If is a folder, we of course don't want to consider this a prefix, but its parent.
-            if !single_file {
-                local_prefix = local_prefix.parent().unwrap();
-            }
+                    // If we are going to compress the local_files we need to take care of the content of
+                    // the .list()-ed files.
+                    // In case of compression of a folder, e.g. if the list_contains glob(/a/folder/**)
+                    // we have to pass the the Remote.upload_folder_compressed only /a/folder for creating
+                    // a single archive.
+                    // Otherwise we'll create a different archive for every file/folder and this is wrong.
+                    let all_with_same_prefix = local_files_clone
+                        .iter()
+                        .all(|path| path.starts_with(local_prefix));
+                    if compress && !single_file && all_with_same_prefix {
+                        single_file = true;
+                        local_files = vec![PathBuf::from(local_prefix)];
+                    }
 
-            // If we are going to compress the local_files we need to take care of the content of
-            // the .list()-ed files.
-            // In case of compression of a folder, e.g. if the list_contains glob(/a/folder/**)
-            // we have to pass the the Remote.upload_folder_compressed only /a/folder for creating
-            // a single archive.
-            // Otherwise we'll create a different archive for every file/folder and this is wrong.
-            let all_with_same_prefix = local_files_clone
-                .iter()
-                .all(|path| path.starts_with(local_prefix));
-            if compress && !single_file && all_with_same_prefix {
-                single_file = true;
-                local_files = vec![PathBuf::from(local_prefix)];
-            }
+                    // Special case in which we want to upload a folder without compression
+                    // If all the files share the same prefix, we upload all the files in this prefix.
+                    // The remote should handle eventual incremental backup.
+                    if !single_file && all_with_same_prefix && !compress {
+                        let remote_path = &remote_prefix;
+                        info!(
+                            "[{}] Uploading a list of files to {}",
+                            name,
+                            remote_path.display()
+                        );
+                        let result = remote.upload_folder(&local_files, remote_path).await;
+                        Backup::log_result(
+                            result,
+                            &name,
+                            local_prefix,
+                            &remote.name(),
+                            remote_path,
+                            compress,
+                        );
+                        info!("[{}] Uploaded completed.", name);
+                        // Set local_files to empty vector for skipping the next loop
+                        // and avoid to add another else branch that will increase the
+                        // indentation again.
+                        local_files = vec![];
+                    }
 
-            // Special case in which we want to upload a folder without compression
-            // If all the files share the same prefix, we upload all the files in this prefix.
-            // The remote should handle eventual incremental backup.
-            if !single_file && all_with_same_prefix && !compress {
-                let remote_path = &remote_prefix;
-                info!(
-                    "[{}] Uploading a list of files to {}",
-                    name,
-                    remote_path.display()
-                );
-                let result = executor::block_on(remote.upload_folder(&local_files, remote_path));
-                log_result(
-                    result,
-                    &name,
-                    local_prefix,
-                    &remote.name(),
-                    remote_path,
-                    compress,
-                );
-                info!("[{}] Uploaded completed.", name);
-                // Set local_files to empty vector for skipping the next loop
-                // and avoid to add another else branch that will increase the
-                // indentation again.
-                local_files = vec![];
-            }
+                    for file in local_files {
+                        let remote_path = if single_file {
+                            remote_prefix.join(file.file_name().unwrap())
+                        } else {
+                            remote_prefix.join(file.strip_prefix(local_prefix).unwrap())
+                        };
 
-            for file in local_files {
-                let remote_path = if single_file {
-                    remote_prefix.join(file.file_name().unwrap())
-                } else {
-                    remote_prefix.join(file.strip_prefix(local_prefix).unwrap())
-                };
+                        let result: Result<(), remote::Error>;
+                        if file.is_dir() {
+                            // compress for sure, the uncompressed scenarios has been treated
+                            // outside this loop
+                            info!(
+                                "[{}] Compressing folder {} and uploading to {}",
+                                name,
+                                file.display(),
+                                remote_path.display()
+                            );
+                            result = remote.upload_folder_compressed(&file, &remote_path).await;
+                        } else if compress {
+                            info!(
+                                "[{}] Compressing file {} and uploading to {}",
+                                name,
+                                file.display(),
+                                remote_path.display()
+                            );
+                            result = remote.upload_file_compressed(&file, &remote_path).await;
+                        } else {
+                            info!(
+                                "[{}] Uploading file {} to {}",
+                                name,
+                                file.display(),
+                                remote_path.display()
+                            );
+                            result = remote.upload_file(&file, &remote_path).await;
+                        }
 
-                let result: Result<(), remote::Error>;
-                if file.is_dir() {
-                    // compress for sure, the uncompressed scenarios has been treated
-                    // outside this loop
-                    info!(
-                        "[{}] Compressing folder {} and uploading to {}",
-                        name,
-                        file.display(),
-                        remote_path.display()
-                    );
-                    result =
-                        executor::block_on(remote.upload_folder_compressed(&file, &remote_path));
-                } else if compress {
-                    info!(
-                        "[{}] Compressing file {} and uploading to {}",
-                        name,
-                        file.display(),
-                        remote_path.display()
-                    );
-                    result = executor::block_on(remote.upload_file_compressed(&file, &remote_path));
-                } else {
-                    info!(
-                        "[{}] Uploading file {} to {}",
-                        name,
-                        file.display(),
-                        remote_path.display()
-                    );
-                    result = executor::block_on(remote.upload_file(&file, &remote_path));
-                }
-
-                // Handle keep_last
-                if let Some(to_keep) = keep_last {
-                    let to_keep = to_keep as usize;
-                    match executor::block_on(remote.enumerate(remote_path.parent().unwrap())) {
-                        Ok(mut list) => {
-                            if list.len() > to_keep {
-                                list.sort();
-                                list.reverse();
-                                for delete_me in &list[to_keep..] {
-                                    if let Some(error) =
-                                        executor::block_on(remote.delete(&PathBuf::from(delete_me)))
-                                            .err()
-                                    {
-                                        error!(
-                                            "[{}] Error during delete of {}: {}",
-                                            name, delete_me, error
-                                        );
-                                    } else {
-                                        info!("[{}] Deleted {}", name, delete_me);
+                        // Handle keep_last
+                        if let Some(to_keep) = keep_last {
+                            let to_keep = to_keep as usize;
+                            match remote.enumerate(remote_path.parent().unwrap()).await {
+                                Ok(mut list) => {
+                                    if list.len() > to_keep {
+                                        list.sort();
+                                        list.reverse();
+                                        for delete_me in &list[to_keep..] {
+                                            if let Some(error) =
+                                                remote.delete(&PathBuf::from(delete_me)).await.err()
+                                            {
+                                                error!(
+                                                    "[{}] Error during delete of {}: {}",
+                                                    name, delete_me, error
+                                                );
+                                            } else {
+                                                info!("[{}] Deleted {}", name, delete_me);
+                                            }
+                                        }
                                     }
                                 }
+                                Err(error) => error!("Error during remote.enumerate: {}", error),
                             }
                         }
-                        Err(error) => error!("Error during remote.enumerate: {}", error),
+
+                        Backup::log_result(
+                            result,
+                            &name,
+                            &file,
+                            &remote.name(),
+                            &remote_path,
+                            compress,
+                        );
                     }
-                }
 
-                log_result(result, &name, &file, &remote.name(), &remote_path, compress);
-            }
-
-            info!(
-                "[{}] Next run: {}",
-                name,
-                schedule.upcoming(chrono::Utc).take(1).next().unwrap()
-            );
-        });
-        scheduler.add(job);
-
-        Ok(())
+                    info!(
+                        "[{}] Next run: {}",
+                        name,
+                        self.schedule.upcoming(chrono::Utc).take(1).next().unwrap()
+                    );
+                })
+            })
+            .unwrap(),
+        )
     }
 }
 
